#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "fftpert.h"

void do_checks1(fftw_complex* data,
                DOUBLE a1, DOUBLE e1, DOUBLE i1,
                DOUBLE a2, DOUBLE e2, DOUBLE i2,
                DOUBLE scale,
                INT part,
                LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM)
{

  LINT i,j,k,l,m;

  LINT indI,indJ,indK,indL,indM;

  LINT ind, cind;

  DOUBLE cur;

  DOUBLE a10max, e10max, e20max, i10max, i20max, a10maxalias;

  DOUBLE sumbessel, sumbessel1;

  DOUBLE aux;

  LINT gcdIK, gcdJL, gcdKM2, gcdLM2, gcdIJKLM2;

     if(a1 != 0.0 && e1 != 0.0 && e2 != 0 && i1 != 0.0 && i2 != 0.0)
       {
        printf("\n\nPerforming check for special values of parameters\n"); 
        printf("   are requested, but none are possible.\n");
       }
     else
       {
        printf("\n\nPerforming checks for special values of parameters:\n\n");

        a10max=0.0;
        e10max=0.0;
        e20max=0.0;
        i10max=0.0;
        i20max=0.0;

        a10maxalias=0.0;

/* compute GCD of relevant couples of the fft boundaries */

/* note that the Diophant equation: a x + b y = c has solutions
   if and only if gcd(a,b) is also a devisor of c */

        gcdIK = gcd(fftI,fftK);
        gcdJL = gcd(fftJ,fftL);
        gcdKM2 = gcd(fftK,fftM/2);
        gcdLM2 = gcd(fftL,fftM/2);
        gcdIJKLM2 = gcd(gcd(gcd(gcd(fftI,fftJ),fftK),fftL),fftM/2);

  for (i=0; i < fftI; i++)
    {
     indI = fftJ * i;
     for (j=0; j < fftJ; j++)
       {
        indJ = fftK * ( j + indI );
        for (k=0; k < fftK; k++)          
	  {
           indK = fftL * ( k + indJ );
           for (l=0; l < fftL; l++)                        
	     {
              indL = (fftM/2+1) * ( l + indK );
              for (m=0; m < fftM/2+1; m++)
		{
                 indM = m + indL;
/* any operations for data[indM] start here */

  cur = fabs(data[indM].re);

/* collecting statistics for the possible tests */

/* alpha = 0 */
  if(a1==0.0 && !(i==0 && k==0 && l==0 && m==0) && cur > a10max)
    {
     a10max = cur;
    }

/* alpha = 0: Bessel functions */
  if(a1==0.0 && i==0 && k==0 && l==0 && m==0)
    {
/* current index with sign */
     if (j<=fftJ/2)
       {
        ind=j;
       }     
     else
       {
        ind=j-fftJ;
       }

/* accumulate maximal error of aliasing */
     sumbessel=besseljn(ind,ind*e2);

     aux=fabs(sumbessel/scale-data[indM].re);

     if(aux > a10maxalias) 
       {
        a10maxalias = aux;
       }

/* accumulate maximal error removing aliasing errors */
     
/* all relevant Bessel functions with indices j+k*fftJ, k>0 */
     cind=ind;
     
     sumbessel1=sumbessel-1.0;

     while(sumbessel1!=sumbessel)
       {
        sumbessel1=sumbessel;
        cind+=fftJ;
        sumbessel+=besseljn(cind,cind*e2);
       }

/* all relevant Bessel functions with indices j+k*fftJ, k<0 */
     cind=ind;

     sumbessel1=sumbessel-1.0;

     while(sumbessel1!=sumbessel)
       {
        sumbessel1=sumbessel;
        cind-=fftJ;
        sumbessel+=besseljn(cind,cind*e2);
       }
      
     aux=fabs(sumbessel/scale-data[indM].re);

     if(aux > a10max) 
       {
        a10max = aux;
       }

    }


/* e1 = 0 */
  if(e1==0.0 && !( (i-k)%gcdIK == 0 ) && cur > e10max)
    {
     e10max = cur;
    }

/* e2 = 0 */
  if(e2==0.0 && !( (j-l)%gcdJL == 0 ) && cur > e20max)
    {
     e20max = cur;
    }

/* i1 = 0 */
  if(i1==0.0 && !( (k-m)%gcdKM2 == 0 ) && cur > i10max)
    {
     i10max = cur;
    }

/* i2 = 0 */
  if(i2==0.0 && !( (m+l)%gcdLM2 == 0 ) && cur > i20max)
    {
     i20max = cur;
    }

/* any operations for data[indM] end here */
		}
	     }
	  }
       }
    }

/* printing messages about maximal violation checks */
  if( a1 == 0.0 )
    {
     printf("alpha = 0 : maximal violation "FORDOUBLE"\n",a10max*scale);
     printf("alpha = 0 : maximal aliasing error "FORDOUBLE"\n",a10maxalias*scale);

        printf(" A_{ijklm}=0 for i!=0 || k!=0 || l!=0 || m!=0\n");
        printf(" A_{0j000}=J_j(j e2)\n\n");

    }

  if( e1 == 0.0 )
    {
     printf("e1 = 0    : maximal violation "FORDOUBLE"\n",e10max*scale);

        printf(" A_{ijklm}=0 for i!=k\n\n");

    }

  if( e2 == 0.0 )
    {
     printf("e2 = 0    : maximal violation "FORDOUBLE"\n",e20max*scale);

        printf(" A_{ijklm}=0 for j!=l\n\n");

    }

  if( i1 == 0.0 )
    {
     printf("i1 = 0    : maximal violation "FORDOUBLE"\n",i10max*scale);

        printf(" A_{ijklm}=0 for k!=m\n\n");

    }  

  if( i2 == 0.0 )
    {
     printf("i2 = 0    : maximal violation "FORDOUBLE"\n",i20max*scale);

        printf(" A_{ijklm}=0 for l!=-m\n\n");

    }

       } 
}







