#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "fftpert.h"

int main(int argc, char *argv[])
{

/* parameters */
  
  DOUBLE a1,a2,e1,e2,i1,i2;
  
  LINT I,J,K,L,M;

  INT direction;

  INT mode;

  INT dalpha,de1,di1,de2,di2;

  INT part;

  INT print_coefficients;

  DOUBLE max_printed;

  INT checks;

/* local auxiliary variables */

  LINT i,j,k,l/*,m*/;

  LINT fftI,fftJ,fftK,fftL,fftM;

  LINT indI,indJ,indK,indL/*,indM*/;

  LINT mem;

/* error estimations and analysis of the result */

  DOUBLE boundary;

  DOUBLE maxcoefficient; 

/* fftw stuff */

  fftw_real *rdata;

  rfftwnd_plan fftwplan;

  fftw_complex *data;

  int *fftwdim;

  const int DIM = 5;

  fftw_real scale;
  
/* time stuff */

  time_t start_table_time, end_table_time, 
         start_plan_time,end_plan_time,
         start_fft_time, end_fft_time,
         start_analyse_time, end_analyse_time;

  clock_t start_table_cputime, end_table_cputime,
          start_plan_cputime,end_plan_cputime, 
          start_fft_cputime, end_fft_cputime,
          start_analyse_cputime, end_analyse_cputime; 

/* actions start here */

/* header */
  
  print_header();

/* reading parameters from the configuration file */

  get_parameters(get_cfg(argc,argv),
                 &a1,&a2,&e1,&e2,&i1,&i2,
                 &I,&J,&K,&L,&M,
                 &direction,
                 &mode,
                 &dalpha,&de1,&di1,&de2,&di2,
                 &part,
                 &print_coefficients,
                 &max_printed,
                 &checks);

/* allocating memory for the values of perturbation function */

  fftI=(I == 0 ? 1 : 2*I);
  fftJ=(J == 0 ? 1 : 2*J);
  fftK=(K == 0 ? 1 : 2*K);
  fftL=(L == 0 ? 1 : 2*L);
  fftM=(M == 0 ? 1 : 2*M);

  printf("Making FFT "FORLINT" *"FORLINT" *"FORLINT" *"FORLINT" *"FORLINT"\n\n",fftI,fftJ,fftK,fftL,fftM);

  mem=fftI*fftJ*fftK*fftL*2*(fftM/2+1)*sizeof(fftw_real);

  printf("Memory need is about %5.0f Mbytes.\n\n", 
         floor(mem/1024.0/1024.0)+1);
 
  rdata = (fftw_real *) malloc(mem);

/* cannot get the memory */

  if ( rdata == NULL )
    { 
     printf("Error: cannot allocate "FORLINT" bytes. Exiting...\n",mem);
     exit(101);
    } 

/* the same data will be accessed as complex after fftw */

  data= (fftw_complex *) rdata;

/* ******************************************************************* */
/* computing table of values of perturbation function */
/* ******************************************************************* */
    
/* timers start */

  time(&start_table_time);
  start_table_cputime = clock();

  printf("Computing table started: %s", ctime(&start_table_time));

/* external routine to compute the table */

  if (dalpha == 0 && de1 == 0 && di1 == 0 && de2 == 0 && di2 == 0)
    {
     if (mode == 1)
       {
        fill_table1(rdata, a1, e1, i1, a2, e2, i2, 
                           fftI, fftJ, fftK, fftL, fftM, 
                           part);
       }
     else if (mode == 2)
       {
        fill_table2(rdata, a1, e1, i1, a2, e2, i2, 
                           fftI, fftJ, fftK, fftL, fftM, 
                           part);
       }
     else if(mode ==3)
       {
        fill_table3(rdata, a1, e1, i1, a2, e2, i2, 
                           fftI, fftJ, fftK, fftL, fftM, 
                           part);
       }
     else
       {
        printf("\n\nUnknown mode requested. Exiting...\n\n");
        exit(1000);
       }
    }
  else if (dalpha == 1 && de1 == 0 && di1 == 0 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(201);
    }
  else if (dalpha == 0 && de1 == 1 && di1 == 0 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(202);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 1 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(203);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 0 && de2 == 1 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(204);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 0 && de2 == 0 && di2 == 1)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(205);
    }
  else if (dalpha == 2 && de1 == 0 && di1 == 0 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(206);
    }
  else if (dalpha == 0 && de1 == 2 && di1 == 0 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(207);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 2 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(208);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 0 && de2 == 2 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(209);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 0 && de2 == 0 && di2 == 2)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(210);
    }
  else if (dalpha == 1 && de1 == 1 && di1 == 0 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(211);
    }
  else if (dalpha == 1 && de1 == 0 && di1 == 1 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(212);
    }
  else if (dalpha == 1 && de1 == 0 && di1 == 0 && de2 == 1 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(213);
    }
  else if (dalpha == 1 && de1 == 0 && di1 == 0 && de2 == 0 && di2 == 1)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(214);
    }
  else if (dalpha == 0 && de1 == 1 && di1 == 1 && de2 == 0 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(215);
    }
  else if (dalpha == 0 && de1 == 1 && di1 == 0 && de2 == 1 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(216);
    }
  else if (dalpha == 0 && de1 == 1 && di1 == 0 && de2 == 0 && di2 == 1)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(217);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 1 && de2 == 1 && di2 == 0)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(218);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 1 && de2 == 0 && di2 == 1)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(219);
    }
  else if (dalpha == 0 && de1 == 0 && di1 == 0 && de2 == 1 && di2 == 1)
    {
     printf("\n\n Sorry. This mode is not yet implemented.\n\n");
     exit(220);
    }
  else
    {
     printf("\n\nYour requested a partial derivative of order more than 2.\n");
     printf("This mode is not implemented. Although it is quite straighforward,\n");
     printf(" we have no intention to implement this mode in the foreseeable\n");
     printf(" future.\n\n");
     exit(221);
    }

/* timers stop */

  time(&end_table_time);
  end_table_cputime = clock();

  printf("Computing table ended: %s", ctime(&end_table_time));

  printf("Elapsed time %10.2f sec.\nCPU time     %10.2f sec.\n\n",
         difftime(end_table_time,start_table_time),
         1.0*(end_table_cputime-start_table_cputime)/CLOCKS_PER_SEC);  

/* ******************************************************************* */
/* creating the fftw plan */
/* ******************************************************************* */

/* timers start */

  time(&start_plan_time);
  start_plan_cputime = clock();

  printf("Creating plan for FFTW: %s", ctime(&start_plan_time));

/* initializing the array of dimensionality */

  fftwdim = (int*) malloc( sizeof(int) * DIM );

  if ( fftwdim == NULL )
    {
     printf("Error: cannot allocate "FORSIZEOF" bytes for fftwdim. Exiting...\n",
            sizeof(int)*DIM);
     exit(21);
    } 

  fftwdim[0] = fftI;
  fftwdim[1] = fftJ;
  fftwdim[2] = fftK;
  fftwdim[3] = fftL;
  fftwdim[4] = fftM;

/* calling theads init if we need multithreaded code */

#ifdef THREADED

   printf("\nInitializing threads for FFTW ...");

   if( fftw_threads_init() != 0 )
     {
/* nonzero answer from fftw_threads_init() mean serious error */
      printf("\n\n FFTW failed to initialize threads. Exiting...\n");
      exit(999);
     }  

   printf(" done.\n\n");
 
#endif

/* calling FFTW routine to create the plan */

  fftwplan = rfftwnd_create_plan(DIM, fftwdim, 
                                 FFTW_REAL_TO_COMPLEX, 
                                 FFTW_MEASURE | FFTW_IN_PLACE);

/* timers stop */

  time(&end_plan_time);
  end_plan_cputime = clock();

  printf("Creating plan ended: %s", ctime(&end_plan_time));

  printf("Elapsed time %10.2f sec.\nCPU time     %10.2f sec.\n\n",
         difftime(end_plan_time,start_plan_time),
         1.0*(end_plan_cputime-start_plan_cputime)/CLOCKS_PER_SEC);  

/* ******************************************************************* */
/* computing FFT */
/* ******************************************************************* */

/* timers start */

  time(&start_fft_time);
  start_fft_cputime = clock();

  printf("Computing FFT: %s", ctime(&start_fft_time));

/* calling fftw routine to compute real-to-complex FFT */

#ifdef THREADED

  rfftwnd_threads_one_real_to_complex(NTHREADS,fftwplan, rdata, NULL);

#else

  rfftwnd_one_real_to_complex(fftwplan, rdata, NULL);

#endif

/* timers stop */

  time(&end_fft_time);
  end_fft_cputime = clock();

  printf("Computing FFT ended: %s", ctime(&end_fft_time));

  printf("Elapsed time %10.2f sec.\nCPU time     %10.2f sec.\n\n",
         difftime(end_fft_time,start_fft_time),
         1.0*(end_fft_cputime-start_fft_cputime)/CLOCKS_PER_SEC);  

/* ******************************************************************* */
/* distroying the fftw plan */
/* ******************************************************************* */

  rfftwnd_destroy_plan(fftwplan); 

/* ******************************************************************* */
/* analysing the result */
/* ******************************************************************* */

/* timers start */

  time(&start_analyse_time);
  start_analyse_cputime = clock();

  printf("Analysis of the result: %s", ctime(&start_analyse_time));

/*---------------------------------------------------------*/
/* scaling the coefficients with m=0 to make them compatible
     with the cosine expansion */
/*---------------------------------------------------------*/

  for (i=0; i < fftI; i++)
    {
     indI = fftJ * i;
     for (j=0; j < fftJ; j++)
       {
        indJ = fftK * ( j + indI );
        for (k=0; k < fftK; k++)          
	  {
           indK = fftL * ( k + indJ );
           for (l=0; l < fftL; l++)            
	     {
              indL = (fftM/2+1) * ( l + indK );

              data[indL].re = 0.5 * data[indL].re;
	     }
	  }
       }
    }


/*---------------------------------------------------------*/
/* computing scaling factor */
/*---------------------------------------------------------*/

/* factor 2.0 is added for compatibility with the cosine expansion */

  scale = 2.0 / ( fftI * fftJ * fftK * fftL * fftM );

  printf("\nScale factor:          %27.16e\n\n",scale);

/*---------------------------------------------------------*/
/* computing boundary for the significant coefficients and maximal coeffient */
/*---------------------------------------------------------*/

set_boundary(data,
             mode,
             scale,
             fftI, fftJ, fftK, fftL, fftM,
             &boundary,
             &maxcoefficient);

/*---------------------------------------------------------*/
/* statistical summary */
/*---------------------------------------------------------*/

statistics(data,
           mode,
           boundary,
           maxcoefficient,
           scale,
           fftI, fftJ, fftK, fftL, fftM);

/*---------------------------------------------------------*/
/* performing checks on request */
/*---------------------------------------------------------*/

  if(checks)
    {
     if( mode == 1) 
       {
        do_checks1(data,
                   a1, e1, i1,
                   a2, e2, i2,
                   scale,
                   part,
                   fftI, fftJ, fftK, fftL, fftM);
       }
     else if (mode == 2)
       {
        do_checks2(data,
                   a1, e1, i1,
                   a2, e2, i2,
                   scale,
                   part,
                   fftI, fftJ, fftK, fftL, fftM);
       }
     else if (mode == 3)
       {
         do_checks3(data,
                    a1, e1, i1,
                    a2, e2, i2,
	            scale,
                    part,
	            fftI, fftJ, fftK, fftL, fftM);
       }
    }

/*---------------------------------------------------------*/
/* printing significant coefficients on request */
/*---------------------------------------------------------*/

  if(print_coefficients)
    {
     print_coeffs(data, 
                  max_printed, boundary,
                  scale,
                  mode,
                  fftI, fftJ, fftK, fftL, fftM);
    }

/* timers stop */

  time(&end_analyse_time);
  end_analyse_cputime = clock();

  printf("\nAnalysis ended: %s", ctime(&end_analyse_time));

  printf("Elapsed time %10.2f sec.\nCPU time     %10.2f sec.\n\n",
         difftime(end_analyse_time,start_analyse_time),
         1.0*(end_analyse_cputime-start_analyse_cputime)/CLOCKS_PER_SEC);  

/* time statistics */

  printf("Total elapsed time %10.2f sec.\nTotal CPU time     %10.2f sec.\n\n",
         difftime(end_analyse_time,start_table_time),
         1.0*(end_analyse_cputime-start_table_cputime)/CLOCKS_PER_SEC);  

/* freeing memory  */

  free(rdata);

/* exiting normally */

  exit(0);

}

