/* fftpert header */

/* with or without threads */

#include "threadsq.h"

/* including real fftw header */

#ifdef THREADED

#include <rfftw_threads.h>

#else

#include <rfftw.h>

#endif

/* version */

#define FFTPERT_VERSION 0.7 /* float */

/* data types */

/* INT is an integer which is always small (e.g., control parameters) */
#define INT int
/* LINT is an integer which could be large (e.g., number of coefficients) */
#define LINT long int
/* DOUBLE is standard float-point number;
   DOUBLE must be equal to fftw_real from the FFTW */
#define DOUBLE double

/* print formats for INT, LINT, DOUBLE and for the result of "sizeof" */

#define FORINT "%d"
#define FORLINT "%ld"
#define FORDOUBLE "%27.16e"
#define FORSIZEOF "%ld"

/* some constants for config file */

#define DEFAULT_CONFIG "fftpert.cfg" /* string */

#define CONFIG_MAX_STRING_LENGTH 1024 /* int */

#define CONFIG_COMMENT_LINE  '#' /* char */

#define CONFIG_DATA_SEPARATOR '=' /* char */

/* mathematical constants */

#define PI 3.141592653589793238462643383279502884197

/* accuracy parameters */

/* internal accuracy of bekep */
/* standard */
/* #define BEKEPEPSLON 1e-15 */
/* maximum for Pentium II with egcs 2.90.29 under Linux */
/* #define BEKEPEPSLON 1e-20 */
/*  maximum for Ultra Sparc with egcs 2.91.60 under Solaris 2.5.1 */
/* #define BEKEPEPSLON 1e-17 */
/*  maximum for Ultra Sparc with cc SC4.0 under Solaris 2.5.1 */
/* #define BEKEPEPSLON 1e-17  */
/*  maximum for MIPS 10000/10010 with egcs 2.91.60 under IRIX 5.6 */
#define BEKEPEPSLON 1e-17
/*  maximum for MIPS 10000/10010 with cc MIPSpro 2.7.1 under IRIX 5.6 */
/* #define BEKEPEPSLON 1e-20 */

/* boundary controls */

#define BOUNDARY_SCALE 2.0

#define BOUNDARY_STEP 10.0

#define BOUNDARY_SAFE 1.4

/* prototypes */

extern FILE* get_cfg(int argc, char **argv);

extern void help();


extern void get_parameters(FILE *cfg,
                    DOUBLE *a1, DOUBLE *a2,
                    DOUBLE *e1, DOUBLE *e2,
                    DOUBLE *i1, DOUBLE *i2,
                    LINT *I, LINT *J, LINT *K, LINT *L, LINT *M, INT *direction,
                    INT *mode,
                    INT *dalpha, INT *de1, INT *de2, INT *di1, INT *di2,
                    INT *part,
                    INT *print_coefficients,
                    DOUBLE *max_printed,
                    INT *checks);

extern void error_in_config(int exit_code, char* curstr);

extern void print_header();

extern double bekep(DOUBLE ex, DOUBLE ey, DOUBLE x0);

extern void fill_table1(fftw_real *rdata,
                       DOUBLE a1, DOUBLE e1, DOUBLE i1,
                       DOUBLE a2, DOUBLE e2, DOUBLE i2,
                       LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM,
		       INT part);

extern void fill_table2(fftw_real *rdata,
                       DOUBLE a1, DOUBLE e1, DOUBLE i1,
                       DOUBLE a2, DOUBLE e2, DOUBLE i2,
                       LINT fftA, LINT fftB, LINT fftC, LINT fftD, LINT fftE,
                       INT part);

extern void fill_table3(fftw_real *rdata,
                       DOUBLE a1, DOUBLE e1, DOUBLE i1,
                       DOUBLE a2, DOUBLE e2, DOUBLE i2,
                       LINT fftA, LINT fftB, LINT fftC, LINT fftD, LINT fftE,
                       INT part);

extern int xcmp(const void *v1, const void *v2);

extern int safeq(INT min, INT max, INT N);

extern INT gcd(INT a, INT b);

extern DOUBLE besselj0(DOUBLE x);

extern DOUBLE besselj1(DOUBLE x);

extern DOUBLE besseljn(INT n, DOUBLE x);

extern void set_boundary(fftw_complex* data,
                         INT mode,
                         DOUBLE scale,
                         LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM,
                         DOUBLE *boundary,
                         DOUBLE *maxcoefficient);

extern void statistics(fftw_complex* data,
                       INT mode,
                       DOUBLE boundary,
                       DOUBLE maxcoefficient,
                       DOUBLE scale,
                       LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM);

extern void do_checks1(fftw_complex* data,
                       DOUBLE a1, DOUBLE e1, DOUBLE i1,
                       DOUBLE a2, DOUBLE e2, DOUBLE i2,
                       DOUBLE scale,
                       INT part,
                       LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM);

extern void do_checks2(fftw_complex* data,
                       DOUBLE a1, DOUBLE e1, DOUBLE i1,
                       DOUBLE a2, DOUBLE e2, DOUBLE i2,
                       DOUBLE scale,
                       INT part,
                       LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM);

extern void do_checks3(fftw_complex* data,
                       DOUBLE a1, DOUBLE e1, DOUBLE i1,
                       DOUBLE a2, DOUBLE e2, DOUBLE i2,
                       DOUBLE scale,
		       INT part,
                       LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM);

extern void print_coeffs(fftw_complex* data,
                  DOUBLE max_printed, DOUBLE boundary,
                  DOUBLE scale,
                  INT mode,
                  LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM);
