#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fftpert.h"

void fill_table1(fftw_real *rdata, 
                DOUBLE a1, DOUBLE e1, DOUBLE i1,
                DOUBLE a2, DOUBLE e2, DOUBLE i2,
                LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM,
                INT part)
{

/* auxiliary table of cos and sin of E1, E2, omega1, omega2, DeltaOmega */

  DOUBLE *cosE1, *sinE1, 
         *cosE2, *sinE2, 
         *cosomega1, *sinomega1, 
         *cosomega2, *sinomega2, 
         *cosDeltaOmega, *sinDeltaOmega;

  DOUBLE arg,step;

  LINT index;

/* auxiliary variable to save time */

  DOUBLE cosi1, sini1, cosi2, sini2, ee1, ee2;

  LINT fftM2p1, twofftM2p1;

/* auxiliary tables of rectangular coordinates */

  DOUBLE *x1, *y1, *z1, *x2, *y2, *z2, *r2;
  LINT indI1, indK1, indM1, indJ2, indL2;

/* auxiliary tables of x1, y1, z1, x2, y2, z2, r2; */

  DOUBLE aux1, aux2, aux3, aux4;
  DOUBLE ccos, csin;

/* computing of the table */

  DOUBLE cx1, cy1, cz1;
  DOUBLE cx2, cy2, cz2, cr2;
  DOUBLE dx, dy, dz2, cz1cz2;

/* local variables */

  LINT i,j,k,l,m;
  LINT indI, indJ, indK, indL, indM;
  LINT indIx, indJx, indKx, indLx, indMx;

/* auxiliary functions of elements to save time later */

  cosi1 = cos(i1);
  sini1 = sin(i1);
  cosi2 = cos(i2);
  sini2 = sin(i2);
  ee1 = sqrt(1.0-e1*e1);
  ee2 = sqrt(1.0-e2*e2);

  fftM2p1 = fftM/2 + 1;
  twofftM2p1 = 2 * fftM2p1;

/* computing auxiliary tables and variables */

  i = fftI * 2 + fftJ * 2 + fftK * 2 + fftI*fftK*(2*fftM2p1+1)+fftJ*(3*fftL+1);
  if (fftL != fftK)
    {
     i += fftL * 2;
    }
  if (fftM != fftK && fftM != fftL)
    {
     i += fftM * 2;
    }
  
printf("\nAuxiliary memory needed temporarily: %5.2f Mbytes\n",
       sizeof(DOUBLE)*i/1024.0/1024.0);

/* a table of cosE and sinE corresponding to e=e1 
   M from 0 to 2Pi with a step 2Pi/fftI */

  cosE1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftI);
  sinE1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftI);


  if ( (cosE1 == NULL) || (sinE1 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for E1. Exiting...\n",
            sizeof(DOUBLE)*fftI*2);
     exit(16);
    } 

  step = 2*PI/fftI; 

  for(index = 0; index < fftI; index++) 
    {
     arg = bekep(e1,0.0,step*index);
     cosE1[index] = cos(arg);
     sinE1[index] = sin(arg);
    }

/* a table of cosE and sinE corresponding to e=e2 
   M from 0 to 2Pi with a step 2Pi/fftJ */

  cosE2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftJ);
  sinE2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftJ);

  if ( (cosE2 == NULL) || (sinE2 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for E2. Exiting...\n",
            sizeof(DOUBLE)*fftJ*2);
     exit(17);
    } 

  step = 2*PI/fftJ; 

  for(index=0; index < fftJ; index++) 
    {
     arg = bekep(e2,0.0,step*index);
     cosE2[index] = cos(arg);
     sinE2[index] = sin(arg);
    }

/* a table of cos omega and sin omega from 0 to 2Pi with a step 2Pi/fftK */

  cosomega1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftK);
  sinomega1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftK);


  if ( (cosomega1 == NULL) || (sinomega1 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for omega1. Exiting...\n",
            sizeof(DOUBLE)*fftK*2);
     exit(18);
    } 
  
  step = 2*PI/fftK; 

  for(index = 0; index < fftK; index++) 
    {
     arg=step*index;
     cosomega1[index] = cos(arg);
     sinomega1[index] = sin(arg);
    }

/* a table of cos omega and sin omega from 0 to 2Pi with a step 2Pi/fftK */

  if ( fftL == fftK)
    {
     cosomega2 = cosomega1;
     sinomega2 = sinomega1;
    }
  else
    {

     cosomega2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftL);
     sinomega2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftL);

     if ( (cosomega2 == NULL) || (sinomega2 == NULL) )
       {
        printf("Error: cannot allocate "FORLINT" bytes for omega2. Exiting...\n",
               sizeof(DOUBLE)*fftL*2);
        exit(18);
       } 

     step = 2*PI/fftL; 

     for(index = 0; index < fftL; index++) 
       {
        arg=step*index;
        cosomega2[index] = cos(arg);
        sinomega2[index] = sin(arg);
       }

    }

/* a table of cos DeltaOmega and sin DeltaOmega from 0 to 2Pi 
   with a step 2Pi/fftM */

  if ( fftM == fftK )
    {
     cosDeltaOmega = cosomega1;
     sinDeltaOmega = sinomega1;
    }
  else if ( fftM == fftL )
    {
     cosDeltaOmega = cosomega2;
     sinDeltaOmega = sinomega2;
    } 
  else
    {

     cosDeltaOmega = (DOUBLE *) malloc(sizeof(DOUBLE)*fftM);
     sinDeltaOmega = (DOUBLE *) malloc(sizeof(DOUBLE)*fftM);

     if ( (cosDeltaOmega == NULL) || (sinDeltaOmega == NULL) )
       {
        printf("Error: cannot allocate "FORLINT" bytes for DeltaOmega. Exiting...\n",
               sizeof(DOUBLE)*fftM*2);
        exit(18);
       } 

     step = 2*PI/fftM; 

     for(index = 0; index < fftM; index++) 
       {
        arg=step*index;
        cosDeltaOmega[index] = cos(arg);
        sinDeltaOmega[index] = sin(arg);
       }

    }

/* auxiliary tables of rectangular coordinates */

/* the first body */

  x1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftI*fftK*fftM2p1);
  y1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftI*fftK*fftM2p1);
  z1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftI*fftK);


  if ( (x1 == NULL) || (y1 == NULL) || (z1 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for x1,y1,z1. Exiting...\n",
            sizeof(DOUBLE)*fftI*fftK*(2*fftM2p1+1));
     exit(19);
    } 
    
  for (i=0; i < fftI; i++)
    {
     indI1 = fftK * i;
     aux1 = -a1 * ee1 * sinE1[i];
     aux2 = a1 * (cosE1[i] - e1); 

     for (k=0; k < fftK; k++)
       {

/* current values of sin and cos of omega */

        csin = sinomega1[k];
        ccos = cosomega1[k];

/* auxiliary variables to save time */

/* x: factor of cos Omega */
        aux3 = aux1 * csin + aux2 * ccos;
/* x: factor of sin Omega */ 
        aux4 = cosi1 * (aux1 * ccos - aux2 * csin);
/* y: factor of cos Omega = -aux4 */
/* y: factor of sin Omega =  aux3 */

/* computing z: independent of Omega */
 
        indK1 = k + indI1;

        z1[indK1] = sini1 * (aux2 * csin - aux1 * ccos );

        indK1 = fftM2p1 * indK1;

/* since we fill up by symmetry m is only to fftM2p1 */
        for (m=0; m < fftM2p1; m++)
	  {

           indM1 = m + indK1;

/* current values of cos and sin of Omega */

           ccos = cosDeltaOmega[m];
           csin = sinDeltaOmega[m];

/* computing x and y */

           x1[indM1] = aux3 * ccos + aux4 * csin; 

           y1[indM1] = aux3 * csin - aux4 * ccos; 

	  }
       }
    }

/* the second body */

  x2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftJ*fftL);
  y2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftJ*fftL);
  z2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftJ*fftL);
  r2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftJ);
  

  if ( (x2 == NULL) || (y2 == NULL) || (z2 == NULL) || (r2 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for x2,y2,z2,r2. Exiting...\n",
            sizeof(DOUBLE)*fftJ*(3*fftL+1));
     exit(20);
    } 

  for (j=0; j < fftJ; j++)
    {

     aux1 = -a2 * ee2 * sinE2[j];
     aux2 = a2 * (cosE2[j] - e2); 

     aux3 = 1.0/(a2 * (1.0 - e2 * cosE2[j]) );

     r2[j] = aux3 * aux3 * aux3 ;

     indJ2 = fftL * j;
  
     for (l=0; l < fftL; l++)
       {
        indL2 = l + indJ2;

/* current values of cos and sin of omega */
        csin = sinomega2[l];
        ccos = cosomega2[l];

        x2[indL2] = aux1 * csin + aux2 * ccos;

        aux4 = (aux2 * csin - aux1 * ccos);

        y2[indL2] = cosi2 * aux4;

        z2[indL2] = sini2 * aux4;

       }
    }

/* initial maximal value of the perturbing function */

  aux1 = 0.0;

/* computing values */

  for (i=0; i < fftI; i++)
    {
     indI = fftJ * i;

/* for picking up x1,y1,z1 */

     indI1 = fftK * i;

/* for filling the table by symmetry */

     indIx = fftJ * ((fftI-i)%fftI);

     for (j=0; j < fftJ; j++)
       {
        indJ = fftK * ( j + indI );

/* for picking up x2,y2,z2,r2 */

        indJ2 = fftL * j;
        cr2 = r2[j];

/* for filling the table by symmetry */

        indJx = fftK * ( ((fftJ-j)%fftJ) + indIx);

        for (k=0; k < fftK; k++)
	  {
           indK = fftL * ( k + indJ );

/* for picking up x1,y1,z1 */

           indK1 = k + indI1;
           cz1 = z1[indK1];
           indK1 = fftM2p1 * indK1;

/* for filling the table by symmetry */

           indKx = fftL * ( ((fftK-k)%fftK) + indJx); 

           for (l=0; l < fftL; l++)
	     {
              indL = twofftM2p1 * ( l + indK );

/* for picking up x2,y2,z2,r2 */

              indL2 = l + indJ2;
              cx2 = x2[indL2];
              cy2 = y2[indL2];
              cz2 = z2[indL2];

/* compute z1-z2 and z1*z2 here to save time */
 
              dz2 = cz1 - cz2;
              dz2 = dz2 * dz2;
              cz1cz2 = cz1 * cz2;

/* for filling the table by symmetry */

              indLx = twofftM2p1 * ( ((fftL-l)%fftL) + indKx );

              for (m=0; m < fftM2p1; m++)
		{
                 indM = m + indL;

/* for filling the table by symmetry */

                 indMx = ( (fftM-m)%fftM ) + indLx;

/* any operations for rdata[indM] start here */

/* for picking up x1,y1 */

                 indM1 = m + indK1;
                 cx1 = x1[indM1];
                 cy1 = y1[indM1]; 

                 dx = cx1 - cx2;
                 dy = cy1 - cy2;

/* taking the requested part: */
/* direct + indirect */
                 if ( part == 0 )
		   {
                    aux2 = 1.0/sqrt(dx*dx+dy*dy+dz2) 
                           -cr2*(cx1*cx2+cy1*cy2+cz1cz2);
		   }
/* direct */
                 else if ( part == 1 )
		   {
                    aux2 = 1.0/sqrt(dx*dx+dy*dy+dz2); 
		   }
/* indirect */
                 else 
		   {
                    aux2 = -cr2*(cx1*cx2+cy1*cy2+cz1cz2);
		   }

                 rdata[indM] = aux2;

/* by symmetry */

       	         rdata[indMx] = aux2;

/* computing maximal value of perturbing function */

                 aux2 = fabs(aux2);
                 
                 if(aux2 > aux1)
		   {
                    aux1 = aux2;
		   }

/* any operations for rdata[indM] end here */
		}
	     }
	  }
       }
    }

/* printing a message about the maximal value */

   printf("\nMaximal value of the perturbing function on current grid: "FORDOUBLE"\n\n",
          aux1);

/* freeing auxiliary arrays for cos and sin of angles */

   free(cosE1);
   free(sinE1);
    
   free(cosE2);
   free(sinE2); 

   free(cosomega1);
   free(sinomega1);
   
   if( fftL != fftK)
     {
      free(cosomega2);
      free(sinomega2);
     }

   if( (fftM != fftK) && (fftM != fftL) )
     {
      free(cosDeltaOmega);
      free(sinDeltaOmega);
     }

/* freeing auxiliary arrays for x1, y1, z1, x2, y2, z2, r2 */

   free(x1);
   free(y1);
   free(z1);
   free(x2);
   free(y2);
   free(z2);
   free(r2);

}















