#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fftpert.h"

void fill_table2(fftw_real *rdata, 
                DOUBLE a1, DOUBLE e1, DOUBLE i1,
                DOUBLE a2, DOUBLE e2, DOUBLE i2,
                LINT fftA, LINT fftB, LINT fftC, LINT fftD, LINT fftE,
                INT part)
{

/* auxiliary table of cos and sin of E1, E2, omega1, omega2, DeltaOmega */

  DOUBLE *cosE1, *sinE1, 
         *cosE2, *sinE2, 
         *cosomega1, *sinomega1, 
         *cosomega2, *sinomega2, 
         *cosDeltaOmega, *sinDeltaOmega;

/* auxiliary variable to save time */

  DOUBLE cosi1, sini1, cosi2, sini2, ee1, ee2, ma1ee1, ma2ee2;

/* auxiliary tables of rectangular coordinates */

  DOUBLE *x1, *y1, *z1, *x2, *y2, *z2, *r2;

  DOUBLE aux1, aux2, aux3, aux4;
  DOUBLE ccos, csin;

/* computing of the table */

  DOUBLE cx1, cy1, cz1;
  DOUBLE cx2, cy2, cz2, cr2;
  DOUBLE dx, dy, dz;

/* local variables */

  LINT a,b,c,d,e;
  LINT indA, indB, indC, indD, indE;
  LINT indAx, indBx, indCx, indDx, indEx;
  LINT indA1, indC1, indE1, indB2, indD2;
  DOUBLE stepa, stepb, stepc, stepd, stepe;
  DOUBLE arga, argb, argc, argd, arge;
  LINT fftE2p1, twofftE2p1;


/* auxiliary functions of elements to save time later */

  cosi1 = cos(i1);
  sini1 = sin(i1);
  cosi2 = cos(i2);
  sini2 = sin(i2);
  ee1 = sqrt(1.0-e1*e1);
  ee2 = sqrt(1.0-e2*e2);
  ma1ee1 = - a1 * ee1;
  ma2ee2 = - a2 * ee2;

  fftE2p1 = fftE/2 + 1;
  twofftE2p1 = 2 * fftE2p1;


/* computing auxiliary tables and variables */

  a = fftA*fftC*2+fftB*fftD*2+fftC*fftE*2+fftD*2+fftA*fftC*fftE2p1*3
      +fftB*fftD*4;
  if (fftE != fftD)
    {
     a += fftE*2;
    }

printf("\nAuxiliary memory needed temporarily: %5.2f Mbytes\n",
       sizeof(DOUBLE)*a/1024.0/1024.0);

/* a table of cosE and sinE corresponding to e=e1 
   M=A-C */

  cosE1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftA*fftC);
  sinE1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftA*fftC);


  if ( (cosE1 == NULL) || (sinE1 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for E1. Exiting...\n",
            sizeof(DOUBLE)*fftA*fftC*2);
     exit(16);
    } 

  stepa = 2*PI/fftA;
  stepc = 2*PI/fftC; 

  for(a = 0; a < fftA; a++) 
    {
     indA = a * fftC;     
     arga = a * stepa;
     for(c = 0; c < fftC; c++)
       {
        indC = c + indA;
        argc = bekep(e1,0.0,arga - c * stepc);
        cosE1[indC] = cos(argc);
        sinE1[indC] = sin(argc);
       }
    }

/* a table of cosE and sinE corresponding to e=e2 
   M=B-D */

  cosE2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftB*fftD);
  sinE2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftB*fftD);


  if ( (cosE2 == NULL) || (sinE2 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for E1. Exiting...\n",
            sizeof(DOUBLE)*fftB*fftD*2);
     exit(16);
    } 

  stepb = 2*PI/fftB;
  stepd = 2*PI/fftD; 

  for(b = 0; b < fftB; b++) 
    {
     indB = b * fftD;     
     argb = b * stepb;
     for(d = 0; d < fftD; d++)
       {
        indD = d + indB;
        argd = bekep(e2,0.0,argb - d * stepd);
        cosE2[indD] = cos(argd);
        sinE2[indD] = sin(argd);
       }
    }

/* a table of cos omega and sin omega from 0 to 2Pi */
/* omega = C-E */


  cosomega1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftC*fftE);
  sinomega1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftC*fftE);


  if ( (cosomega1 == NULL) || (sinomega1 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for E1. Exiting...\n",
            sizeof(DOUBLE)*fftC*fftE*2);
     exit(16);
    } 

  stepc = 2*PI/fftC;
  stepe = 2*PI/fftE; 

  for(c = 0; c < fftC; c++) 
    {
     indC = c * fftE;     
     argc = c * stepc;
     for(e = 0; e < fftE; e++)
       {
        indE = e + indC;
        arge = argc - e * stepe;
        cosomega1[indE] = cos(arge);
        sinomega1[indE] = sin(arge);
       }
    }

/* a table of cos omega' and sin omega' from 0 to 2Pi with a step 2Pi/fftD */

  cosomega2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftD);
  sinomega2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftD);

  if ( (cosomega2 == NULL) || (sinomega2 == NULL) )
     {
      printf("Error: cannot allocate "FORLINT" bytes for omega2. Exiting...\n",
             sizeof(DOUBLE)*fftD*2);
      exit(18);
     } 

  stepd = 2*PI/fftD; 

  for(d = 0; d < fftD; d++) 
     {
      argd = d * stepd;
      cosomega2[d] = cos(argd);
      sinomega2[d] = sin(argd);
     }

/* a table of cos DeltaOmega and sin DeltaOmega from 0 to 2Pi 
   with a step 2Pi/fftM */

  if ( fftE == fftD )
    {
     cosDeltaOmega = cosomega2;
     sinDeltaOmega = sinomega2;
    } 
  else
    {

     cosDeltaOmega = (DOUBLE *) malloc(sizeof(DOUBLE)*fftE);
     sinDeltaOmega = (DOUBLE *) malloc(sizeof(DOUBLE)*fftE);

     if ( (cosDeltaOmega == NULL) || (sinDeltaOmega == NULL) )
       {
        printf("Error: cannot allocate "FORLINT" bytes for DeltaOmega. Exiting...\n",
               sizeof(DOUBLE)*fftE*2);
        exit(18);
       } 

     stepe = 2*PI/fftE; 

     for(e = 0; e < fftE; e++) 
       {
        arge = stepe * e;
        cosDeltaOmega[e] = cos(arge);
        sinDeltaOmega[e] = sin(arge);
       }

    }

/* auxiliary tables of rectangular coordinates */

/* the first body */

  x1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftA*fftC*fftE2p1);
  y1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftA*fftC*fftE2p1);
  z1 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftA*fftC*fftE2p1);


  if ( (x1 == NULL) || (y1 == NULL) || (z1 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for x1,y1,z1. Exiting...\n",
            sizeof(DOUBLE)*fftA*fftC*fftE2p1*3);
     exit(19);
    } 

  for (a=0; a < fftA; a++)
    {
     indA = fftC * a;

     for (c=0; c < fftC; c++)
       {
        indC = c + indA;

        aux1 = ma1ee1 * sinE1[indC];
        aux2 = a1 * (cosE1[indC] - e1); 

        indC = fftE2p1 * indC;

        indC1 = c * fftE;

        for (e=0; e <fftE2p1; e++)
	  {
           indE = e + indC;

           indE1 = e + indC1;
           csin = sinomega1[indE1];
           ccos = cosomega1[indE1];

/* auxiliary variables to save time */

/* x: factor of cos Omega */
           aux3 = aux1 * csin + aux2 * ccos;
/* x: factor of sin Omega */ 
           aux4 = /* cosi1 * (  ) */ aux1 * ccos - aux2 * csin;
/* y: factor of cos Omega = -aux4 */
/* y: factor of sin Omega =  aux3 */

/* computing z: independent of Omega */

           z1[indE] = -sini1 * aux4;

           aux4 = cosi1 * aux4;

/* current values of cos and sin of Omega */

           ccos = cosDeltaOmega[e];
           csin = sinDeltaOmega[e];

/* computing x and y */

           x1[indE] = aux3 * ccos + aux4 * csin; 

           y1[indE] = aux3 * csin - aux4 * ccos; 

	  }
       }
    }

/* the second body */

  x2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftB*fftD);
  y2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftB*fftD);
  z2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftB*fftD);
  r2 = (DOUBLE *) malloc(sizeof(DOUBLE)*fftB*fftD);
  

  if ( (x2 == NULL) || (y2 == NULL) || (z2 == NULL) || (r2 == NULL) )
    {
     printf("Error: cannot allocate "FORLINT" bytes for x2,y2,z2,r2. Exiting...\n",
            sizeof(DOUBLE)*fftB*fftD*4);
     exit(20);
    } 

  for (b=0; b < fftB; b++)
    {
     indB = b * fftD;

     for (d=0; d < fftD; d++)
       {
        indD = d + indB;
 
        aux1 = ma2ee2 * sinE2[indD];
        aux2 = a2 * (cosE2[indD] - e2);

        aux3 = 1.0/(a2 * (1.0 - e2 * cosE2[indD]) );

        r2[indD] = aux3 * aux3 * aux3 ;

/* current values of cos and sin of omega */
        csin = sinomega2[d];
        ccos = cosomega2[d];

        x2[indD] = aux1 * csin + aux2 * ccos;

        aux4 = aux2 * csin - aux1 * ccos;

        y2[indD] = cosi2 * aux4;

        z2[indD] = sini2 * aux4;

       }
    }

/* initial maximal value of the perturbing function */

  aux1 = 0.0;

/* computing values */

  for (a=0; a < fftA; a++)
    {
     indA = fftB * a;

/* filling up by symmetry */

     indAx = fftB * ((fftA-a)%fftA);
     
/* picking up x1,y1,z1 */
     
     indA1 = fftC * a;

     for (b=0; b < fftB; b++)
       {
        indB = fftC * ( b + indA );

/* filling up by symmetry */

        indBx = fftC * ( ((fftB-b)%fftB) + indAx);

/* picking up x2,y2,z2,r2 */

        indB2 = fftD * b;

        for (c=0; c < fftC; c++)
	  {
           indC = fftD * ( c + indB );

/* filling up by symmetry */

           indCx = fftD * ( ((fftC-c)%fftC) + indBx);

/* picking up x1,y1,z1 */
           
           indC1 = fftE2p1 * ( c + indA1 );

           for (d=0; d < fftD; d++)
	     {
              indD = twofftE2p1 * ( d + indC );
     
/* filling up by symmetry */

              indDx = twofftE2p1 * ( ((fftD-d)%fftD) + indCx );
         
/* picking up x2,y2,z2,r2 */

              indD2 = d + indB2;
              cr2 = r2[indD2];
              cx2 = x2[indD2];
              cy2 = y2[indD2];
              cz2 = z2[indD2];

              for (e=0; e < fftE2p1; e++)
		{
                 indE = e + indD;

/* any operations for rdata[indE] start here */

/* filling up by symmetry */

                 indEx = ( (fftE-e)%fftE ) + indDx;

/* for picking up x1,y1,z1 */

                 indE1 = e + indC1;

                 cx1 = x1[indE1];
                 cy1 = y1[indE1]; 
                 cz1 = z1[indE1];

                 dx = cx1 - cx2;
                 dy = cy1 - cy2;
                 dz = cz1 - cz2;

/* taking the requested part: */
/* direct + indirect */
                 if ( part == 0)
		   {
                    aux2 = 1.0/sqrt(dx*dx+dy*dy+dz*dz) 
                           -cr2*(cx1*cx2+cy1*cy2+cz1*cz2);
		   }
/* direct */
                 else if (part == 1)
		   {
                    aux2 = 1.0/sqrt(dx*dx+dy*dy+dz*dz); 
		   }
/* indirect */
                 else 
		   {
                    aux2 = -cr2*(cx1*cx2+cy1*cy2+cz1*cz2);
		   }

                 rdata[indE] = aux2;

                 rdata[indEx] = aux2;

/* computing maximal value of perturbing function */

                 aux2 = fabs(aux2);
                 
                 if(aux2 > aux1)
		   {
                    aux1 = aux2;
		   }

/* any operations for rdata[indE] end here */
		}
	     }
	  }
       }
    }

/* printing a message about the maximal value */

   printf("\nMaximal value of the perturbing function on current grid: "FORDOUBLE"\n\n",
          aux1);


/* freeing auxiliary arrays for cos and sin of angles */

   free(cosE1);
   free(sinE1);
    
   free(cosE2);
   free(sinE2); 

   free(cosomega1);
   free(sinomega1);
   
   free(cosomega2);
   free(sinomega2);

   if( fftE != fftE )
     {
      free(cosDeltaOmega);
      free(sinDeltaOmega);
     }

/* freeing auxiliary arrays for x1, y1, z1, x2, y2, z2, r2 */

   free(x1);
   free(y1);
   free(z1);
   free(x2);
   free(y2);
   free(z2);
   free(r2);

}















