#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fftpert.h"

void fill_table3(fftw_real *rdata, 
                DOUBLE a1, DOUBLE e1, DOUBLE i1,
                DOUBLE a2, DOUBLE e2, DOUBLE i2,
                LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM,
                INT part)
{

  DOUBLE maxpert;

/* local variables */

  LINT i,j,k,l,m;
  LINT indI, indJ, indK, indL, indM;
  LINT indIx, indJx, indKx, indLx, indMx;

  DOUBLE E1,E2,om1,om2,dOM;

  DOUBLE cE1, sE1, cE2, sE2, com1, som1, com2, som2, cdOM, sdOM;

  DOUBLE x1,y1,z1,x2,y2,z2,r2m3,dx,dy,dz,dz2dz;

  DOUBLE aux11,aux12,aux21,aux22,aux31,aux41,aux;

/* auxiliary variable to save time */

  DOUBLE cosi1, sini1, cosi2, sini2, ee1, ee2, ma1ee1, ma2ee2;

  LINT fftM2p1, twofftM2p1;

  DOUBLE stepI, stepJ, stepK, stepL, stepM;

/* auxiliary functions of elements to save time later */

  cosi1 = cos(i1);
  sini1 = sin(i1);
  cosi2 = cos(i2);
  sini2 = sin(i2);
  ee1 = sqrt(1.0-e1*e1);
  ee2 = sqrt(1.0-e2*e2);
  ma1ee1 = - a1 * ee1;
  ma2ee2 = - a2 * ee2;

  fftM2p1 = fftM/2 + 1;
  twofftM2p1 = 2 * fftM2p1;

/* steps for each dimension */

  stepI = 2*PI/fftI;
  stepJ = 2*PI/fftJ;
  stepK = 2*PI/fftK;
  stepL = 2*PI/fftL;
  stepM = 2*PI/fftM;

/* initial maximal value of the perturbing function */

  maxpert = 0.0;

/* computing values */

  for (i=0; i < fftI; i++)
    {
     indI = fftJ * i;

/* for filling the table by symmetry */

     indIx = fftJ * ((fftI-i)%fftI);

/* compute cos(E1), sin(E1) here */

     E1=bekep(e1,0.0,stepI*i);

     cE1=cos(E1);
     sE1=sin(E1);

     aux11 = ma1ee1 * sE1;
     aux21 = a1 * (cE1 - e1);


     for (j=0; j < fftJ; j++)
       {
        indJ = fftK * ( j + indI );

/* for filling the table by symmetry */

        indJx = fftK * ( ((fftJ-j)%fftJ) + indIx);

/* compute cos(E2), sin(E2) and r2m3=r2^(-3) here */

              E2=bekep(e2,0.0,stepJ*j);

              cE2 = cos(E2);
              sE2 = sin(E2);

              aux12 = ma2ee2 * sE2;
              aux22 = a2 * (cE2 - e2);

              aux = 1.0/(a2 * (1.0 - e2 * cE2) );

              r2m3 = aux * aux * aux ;

        for (k=0; k < fftK; k++)
	  {
           indK = fftL * ( k + indJ );

/* for filling the table by symmetry */

           indKx = fftL * ( ((fftK-k)%fftK) + indJx); 

/* current values of cos and sin of omega_1 and z1 */

           om1=-stepI*i+stepK*k;
           som1 = sin(om1);
           com1 = cos(om1);

/* x: factor of cos Omega */
           aux31 = aux11 * som1 + aux21 * com1;
/* x: factor of sin Omega */
           z1 = aux21 * som1 - aux11 * com1; 
           aux41 = -cosi1 * z1;
           z1 = sini1 * z1;
/* y: factor of cos Omega = -aux41 */
/* y: factor of sin Omega =  aux31 */

           for (l=0; l < fftL; l++)
	     {
              indL = twofftM2p1 * ( l + indK );

/* for filling the table by symmetry */

              indLx = twofftM2p1 * ( ((fftL-l)%fftL) + indKx );

/* compute x2,y2,z2 here */

/* current values of cos and sin of omega_2 */

              om2=-stepJ*j+stepL*l;
              som2 = sin(om2);
              com2 = cos(om2);

              x2 = aux12 * som2 + aux22 * com2;

              aux = aux22 * som2 - aux12 * com2;

              y2 = cosi2 * aux;

              z2 = sini2 * aux;

/* z distance between the two bodies */

              dz = z2-z1;

              dz2dz=dz*dz;

              for (m=0; m < fftM2p1; m++)
		{
                 indM = m + indL;

/* for filling the table by symmetry */

                 indMx = ( (fftM-m)%fftM ) + indLx;

/* x and y coordinates of the first body */

                 dOM = -stepK*k+stepL*l+stepM*m;
                 sdOM = sin(dOM);
                 cdOM = cos(dOM);

                 x1 = aux31 * cdOM + aux41 * sdOM; 
                 y1 = aux31 * sdOM - aux41 * cdOM; 

/* x and y distance between the two bodies */

                 dx=x1-x2;
                 dy=y1-y2;

/* taking the requested part: */
/* direct + indirect */
                 if ( part == 0 )
		   {
                    aux = 1.0/sqrt(dx*dx+dy*dy+dz2dz) 
                          -r2m3*(x1*x2+y1*y2+z1*z2);
		   }
/* direct */
                 else if ( part == 1 )
		   {
                    aux = 1.0/sqrt(dx*dx+dy*dy+dz2dz); 
		   }
/* indirect */
                 else 
		   {
                    aux = -r2m3*(x1*x2+y1*y2+z1*z2);
		   }

                 rdata[indM] = aux;

/* by symmetry */

       	         rdata[indMx] = aux;

/* computing maximal value of perturbing function */

                 aux = fabs(aux);
                 
                 if(aux > maxpert)
		   {
                    maxpert = aux;
		   }

/* any operations for rdata[indM] end here */
		}
	     }
	  }
       }
    }

/* printing a message about the maximal value */

   printf("\nMaximal value of the perturbing function on current grid: "FORDOUBLE"\n\n",
          maxpert);

}







