#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "fftpert.h"


void get_parameters(FILE *cfg, 
                    DOUBLE *a1, DOUBLE *a2, 
                    DOUBLE *e1, DOUBLE *e2,
                    DOUBLE *i1, DOUBLE *i2,
                    LINT *I, LINT *J, LINT *K, LINT *L, LINT *M, INT *direction,
                    INT *mode,
                    INT *dalpha, INT *de1, INT *de2, INT *di1, INT *di2,
                    INT *part,
                    INT *print_coefficients,
                    DOUBLE *max_printed,
                    INT *checks)
{

   char *loc, *curstr;
   INT i;
   DOUBLE aux;

   curstr = (char*) malloc( sizeof(char) * CONFIG_MAX_STRING_LENGTH );

   for (i=1; i<23; i++)
     {

/* get next non-comment line */

      if( fgets( curstr, CONFIG_MAX_STRING_LENGTH, cfg) == NULL )
        {
         printf("Error in config file. Unexpected end of file.\n");
         exit(31);
        }
   
      while ( curstr[0] == CONFIG_COMMENT_LINE )
        {
         if( fgets( curstr, CONFIG_MAX_STRING_LENGTH, cfg) == NULL )
           {
            printf("Error in config file. Unexpected end of file.\n");
            exit(32);
           }
        }
   
/* get rid of comments before CONFIG_DATA_SEPARATOR */
   
      if ( (loc = strchr(curstr, CONFIG_DATA_SEPARATOR)) == NULL )
        { 
         error_in_config(4,curstr);
        }
      else
        {
         loc = loc + sizeof(char);   
/* transform and check parameters */         
         switch (i)
	   {
	    case 1:
	       {
                *a1 = atof(loc);
                if ( *a1 < 0.0 ) error_in_config(5,curstr);
                break;
               } 
	    case 2:
	       {
                *e1 = atof(loc);
                if ( (*e1 < 0.0) | (*e1 >= 1.0) ) error_in_config(6,curstr);
                break;
               } 
	    case 3:
	       {
                *i1 = atof(loc);
                if ( (*i1 < -90.0) | (*i1 > 90.0) ) error_in_config(7,curstr);
                break;
               }  
	    case 4:
	       {
                *a2 = atof(loc);
                if ( *a2 < 0.0 ) error_in_config(8,curstr);
                break;
               } 
 	    case 5:
	       {
                *e2 = atof(loc);
                if ( (*e2 < 0.0) | (*e2 >= 1.0) ) error_in_config(9,curstr);
                break;
               } 
 	    case 6:
	       {
                *i2 = atof(loc);
                if ( (*i2 < -90.0) | (*i2 > 90.0) ) error_in_config(10,curstr);
                break;
               } 
	    case 7:
	       {
                *I = atoi(loc);
                if ( *I < 0 ) error_in_config(11,curstr);
                break;
               } 
	    case 8:
	       {
                *J = atoi(loc);
                if ( *J < 0 ) error_in_config(11,curstr);
                break;
               } 
	    case 9:
	       {
                *K = atoi(loc);
                if ( *K < 0 ) error_in_config(12,curstr);
                break;
               } 
	    case 10:
	       {
                *L = atoi(loc);
                if ( *L < 0 ) error_in_config(13,curstr);
                break;
               } 
	    case 11:
	       {
                *M = atoi(loc);
                if ( *M < 0 ) error_in_config(14,curstr);
                break;
               } 
	    case 12:
	       {
                *direction = atoi(loc);
                if ( (*direction != 1) & (*direction != 2) ) error_in_config(15,curstr);
                break;
               } 	    
            case 13:
	       {
                *mode = atoi(loc);
                if ( (*mode <1) | (*mode >3) ) error_in_config(16,curstr);
                break;
               } 
	    case 14:
	      {
               *dalpha = atoi(loc);
               if ( *dalpha < 0 ) error_in_config(17,curstr);
               break;
	      }
	    case 15:
	      {
               *de1 = atoi(loc);
               if ( *de1 < 0 ) error_in_config(18,curstr);
               break;
	      }    
	    case 16:
	      {
               *di1 = atoi(loc);
               if ( *di1 < 0 ) error_in_config(19,curstr);
               break;
	      }
	    case 17:
	      {
               *de2 = atoi(loc);
               if ( *de2 < 0 ) error_in_config(20,curstr);
               break;
	      }
	    case 18:
	      {
               *di2 = atoi(loc);
               if ( *di2 < 0 ) error_in_config(21,curstr);
               break;
	      }
	    case 19:
	      {
               *part = atoi(loc);
               if ( (*part < 0) | (*part > 2) ) error_in_config(22,curstr);
               break;
	      }
	    case 20:
	       {
                *print_coefficients = atoi(loc);
                break;
               } 
	    case 21:
	       {
                *max_printed = atof(loc);
                if ( *max_printed < 0 ) error_in_config(23,curstr);
                break;
               } 
	    case 22:
	       {
                *checks = atoi(loc);
                break;
               } 
	   }
        }

     }

/* close the config file */

   fclose(cfg);

/* print parameters */

   printf("Parameters:\n\n");
   
   printf("first semi-major axis       : "FORDOUBLE"\n",*a1);
   printf("first eccentricity          : "FORDOUBLE"\n",*e1);
   printf("first inclination           : "FORDOUBLE"\n",*i1);
   *i1=*i1*PI/180.0;
   printf("      sin(i)            : "FORDOUBLE"\n",sin(*i1));
   printf("      sin(i/2)          : "FORDOUBLE"\n\n",sin(*i1/2));


   printf("second semi-major axis      : "FORDOUBLE"\n",*a2);
   printf("second eccentricity         : "FORDOUBLE"\n",*e2);
   printf("second inclination          : "FORDOUBLE"\n",*i2);
   *i2=*i2*PI/180.0;
   printf("      sin(i)            : "FORDOUBLE"\n",sin(*i2));
   printf("      sin(i/2)          : "FORDOUBLE"\n\n",sin(*i2/2));


   printf("alpha=min(a1,a2)/max(a1,a2) : "FORDOUBLE"\n\n",
          ( *a1 < *a2 ) ? *a1 / *a2 : *a2 / *a1);

   printf("range of the 1st parameter           ->  "FORLINT" : "FORLINT"\n",
          ( *I == 0 ) ? 0: -(*I)+1, *I);   
   printf("range of the 2nd parameter           ->  "FORLINT" : "FORLINT"\n",
          ( *J == 0 ) ? 0: -(*J)+1, *J);  
   printf("range of the 3rd parameter           ->  "FORLINT" : "FORLINT"\n",
          ( *K == 0 ) ? 0: -(*K)+1, *K);  
   printf("range of the 4th parameter           ->  "FORLINT" : "FORLINT"\n",
          ( *L == 0 ) ? 0: -(*L)+1, *L);
   printf("range of the 5th parameter           ->  "FORLINT" : "FORLINT"\n\n",
          (LINT) 0, *M);

  if( *part == 0 )
    {
     printf("The whole ");
    }
  else if ( *part == 1 )
    {
     printf("The direct part of the ");
    }
  else
    {
     printf("The indirect part of the "); 
    }

  printf("disturbing function for the"); 

  if ( *direction == 1 )
    {
      printf(" first ");
    }
  else
    {
      printf(" second ");
 
/* fftpert knows nothing about the requested direction */
/*   to this end we swap the parameters here */
     
      aux = *a1;
      *a1 = *a2;
      *a2 = aux;

      aux = *i1;
      *i1 = *i2;
      *i2 = aux;

      aux = *e1;
      *e1 = *e2;
      *e2 = aux;

    }

  printf("planet will be computed.\n");

  /* scaling of the semi-major axes: R = Gm'/a' * (what we compute) */ 

  if(*a2 == 0.0)
    {
     printf("\nWrong parameters: the acting planet must have a positive semi-major axis...\n");     
     exit(30);
    }
  else
    {
     *a1 = (*a1) / (*a2);  
     *a2 = 1.0;
    }


/* printing info on the requested mode */

  printf("Requested arguments: ");
  if(*mode ==1)
    {
     printf("M, M', omega, omega', Omega-Omega'\n\n");
    }
  else if(*mode == 2)
    {
     printf("M+omega+Omega-Omega', M'+omega', omega+Omega-Omega', omega', Omega-Omega'\n\n");
    }
  else 
    {
     printf("M, M', M+omega, M'+omega', M+omega+Omega-M'-omega'-Omega'\n\n");
    }

  printf("Requested function:\n");
  printf("  order of partial derivative with respect to alpha : "FORINT"\n",*dalpha);
  printf("  order of partial derivative with respect to e1    : "FORINT"\n",*de1);
  printf("  order of partial derivative with respect to i1    : "FORINT"\n",*di1);
  printf("  order of partial derivative with respect to e2    : "FORINT"\n",*de2);
  printf("  order of partial derivative with respect to i2    : "FORINT"\n\n",*di2);

/* auxiliary parameters */

  if(*print_coefficients)
    {
     if (*max_printed == 0.0e0)
       {
        printf("\nSignificant coefficients will be printed.\n");
       }
     else
       {
        printf("\nThe coefficients which are larger than "FORDOUBLE" will be printed.\n",*max_printed);
       }
    }
  else
    {
     printf("No coefficients will be printed.\n");
    }

  if(*checks)
    {
     printf("All known checks for particular values of elements will be done.\n\n");
    }
  else
    {
     printf("No checks for particular values of elements are requested.\n\n");
    }
 
}  










