#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "fftpert.h"

void set_boundary(fftw_complex* data,
                  INT mode,
                  DOUBLE scale,
                  LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM,
                  DOUBLE *boundary,
                  DOUBLE *maxcoefficient)
{

/* error estimations and analysis of the result */

  DOUBLE maximaginary,cur,maxparity,maxerror0,maxerror2;

  LINT indIx,indJx,indKx,indLx,indMx;

  LINT indI,indJ,indK,indL,indM;

  LINT i,j,k,l,m;

  DOUBLE *boundaries;

  LINT mem;
  
  LINT ind; 

/*---------------------------------------------------------*/
/* estimating imaginary parts and maximal coefficient: 
     linear search is faster */
/*---------------------------------------------------------*/

  maximaginary=0.0;

  *maxcoefficient=0.0;

  mem = fftI * fftJ * fftK * fftL * (fftM/2 +1);

  for(ind=0; ind < mem; ind++)
    {

     cur = fabs(data[ind].im);

     if ( cur > maximaginary ) 
       {
        maximaginary = cur;
       }

     cur = fabs(data[ind].re);

     if ( cur > (*maxcoefficient) ) 
       {
        *maxcoefficient = cur;
       }
     
    }

/* statistical message about maximal imaginary part */

printf("Maximal imaginary part "FORDOUBLE"\n\n", maximaginary * scale);       

/* statistical message about maximal coefficient */

printf("Maximal coefficient    "FORDOUBLE"\n\n", (*maxcoefficient) * scale);       

/*---------------------------------------------------------*/
/* check for the symmetries for m=0 and m=fftM/2 */
/*---------------------------------------------------------*/

maxerror0 = 0.0;

maxerror2 = 0.0;

  for (i=0; i < fftI/2+1; i++)
    {
     indI = fftJ * i;
     indIx = fftJ * ((fftI - i)%fftI);
     for (j=0; j < fftJ/2+1; j++)
       {
        indJ = fftK * ( j + indI );
        indJx = fftK * ( ((fftJ - j)%fftJ) + indIx );
        for (k=0; k < fftK/2+1; k++)          
	  {
           indK = fftL * ( k + indJ );
           indKx = fftL * ( ((fftK - k)%fftK) + indJx );
           for (l=0; l < fftL/2+1; l++)            
	     {
              indL = (fftM/2+1) * ( l + indK );
              indLx = (fftM/2+1) * ( ((fftL - l)%fftL) + indKx );

              indM = indL;
              indMx = indLx;

/* any operations for data[indM] start here */

     cur = fabs(data[indM].re-data[indMx].re); 

     if ( cur > maxerror0 ) 
       {
        maxerror0 = cur;
       }

/* any operations for data[indM] end here */

              indM = fftM/2 + indL;
              indMx = fftM/2 + indLx;

/* any operations for data[indM] start here */

     cur = fabs(data[indM].re-data[indMx].re); 

     if ( cur > maxerror2 ) 
       {
        maxerror2 = cur;
       }

/* any operations for data[indM] end here */
		
	     }
	  }
       }
    }

/* statistical message about maximal difference between
   Coef(i,j,k,l,m) and Coef(fftI-i,fftJ-j,fftK-k,fftL-l,m)
   for m=0 and m=fftM/2 */

printf("Maximal difference Coef(i)-Coef(I-i) for m=0:      "FORDOUBLE"\n",
       maxerror0 * scale);

printf("Maximal difference Coef(i)-Coef(I-i) for m=fftM/2: "FORDOUBLE"\n\n",
       maxerror2 * scale);

/*---------------------------------------------------------*/
/* maximal coeffients with different parities of k and l */
/*---------------------------------------------------------*/

maxparity = 0.0;

  for (i=0; i < fftI; i++)
    {
     indI = fftJ * i;
     for (j=0; j < fftJ; j++)
       {
        indJ = fftK * ( j + indI );
        for (k=0; k < fftK; k++)          
	  {
           indK = fftL * ( k + indJ );
/* here we want to consider l with parity different as that of k (mode==1)*/
/* for mode==2 l should be NOT of the same parity as i+j+k */
/* for mode==3 l should be NOT of the same parity as k */
           for (l=1 - ( (mode == 2) ? (i+j+k)%2 : k%2 ); l < fftL; l=l+2)     
	     {
              indL = (fftM/2+1) * ( l + indK );
              for (m=0; m < fftM/2+1; m++)
		{
                 indM = m + indL;

/* any operations for data[indM] start here */

     cur = fabs(data[indM].re);

     if ( cur > maxparity ) 
       {
        maxparity = cur;
       }

/* any operations for data[indM] end here */
		}
	     }
	  }
       }
    }

/* statistical message about maximal real coefficient with different parity 
   of k and l */

printf("Maximal different parity coefficient  "FORDOUBLE"\n", maxparity * scale);

/*---------------------------------------------------------*/
/* setting boundary for the significant coefficients */
/*---------------------------------------------------------*/

/* set boundary: maximum of maximaginary, maxparity, maxerror0, maxerror2 
   times BOUNDARY_SCALE factor defined in fftpert.h */
  
  boundaries = (DOUBLE *) malloc(4*sizeof(DOUBLE));

  boundaries[0]=maximaginary;
  boundaries[1]=maxparity;
  boundaries[2]=maxerror0;
  boundaries[3]=maxerror2;

  qsort(boundaries,4,sizeof(DOUBLE),xcmp);

  *boundary = boundaries[0] * BOUNDARY_SCALE;

  free(boundaries);

  printf("\nBoundary of significant coefficients is set to be "FORDOUBLE"\n",
         (*boundary) * scale);


}
