#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "fftpert.h"

void statistics(fftw_complex* data,
                INT mode,
                DOUBLE boundary,
                DOUBLE maxcoefficient,
                DOUBLE scale,
                LINT fftI, LINT fftJ, LINT fftK, LINT fftL, LINT fftM)
{

  DOUBLE n1,n2;
 
  INT nbounds;

  LINT ncoeffs;

  LINT *counters;

  LINT ci,cj,ck,cl,cm;

  LINT i,j,k,l,m;

  LINT *imins,*imaxs,*jmins,*jmaxs,*kmins,*kmaxs,*lmins,*lmaxs,*mmins,*mmaxs;

  DOUBLE *boundaries;

  LINT indI,indJ,indK,indL,indM;

  DOUBLE cur;
  
  LINT ind, cind;

/* create a set of boundaries and counters */

/* computing the number of value-boundaries to be used:
   the lowest boundary is always the boundary computed above;
   the highest boundary is always the maximal coefficient found above;
   all other are BOUNDARY_STEP**n and between the lowest and 
     the highest boundaries;
   BOUNDARY_STEP is defined in fftpert.h */

  n1 =  ceil( log(boundary*scale) / log(BOUNDARY_STEP) );
  n2 =  floor( log(maxcoefficient*scale) / log(BOUNDARY_STEP) );
  nbounds = (int) n2 - (int) n1 + 1;
  if(nbounds < 1) 
    {
     nbounds = 2;
    }
  else
    {
     nbounds += 2;
    }    

/* allocating memory for counters, index-boundaries and value-boundaries */

  counters= (LINT *) malloc(nbounds*sizeof(LINT));

  imins = (LINT *) malloc(nbounds*sizeof(LINT));
  imaxs = (LINT *) malloc(nbounds*sizeof(LINT));
  jmins = (LINT *) malloc(nbounds*sizeof(LINT));
  jmaxs = (LINT *) malloc(nbounds*sizeof(LINT));
  kmins = (LINT *) malloc(nbounds*sizeof(LINT));
  kmaxs = (LINT *) malloc(nbounds*sizeof(LINT));
  lmins = (LINT *) malloc(nbounds*sizeof(LINT));
  lmaxs = (LINT *) malloc(nbounds*sizeof(LINT));
  mmins = (LINT *) malloc(nbounds*sizeof(LINT));
  mmaxs = (LINT *) malloc(nbounds*sizeof(LINT));

  boundaries = (DOUBLE *) malloc(nbounds*sizeof(DOUBLE));

/* initiating the counters and index-boundaries */

  for(ind=0; ind < nbounds; ind++)
    {
     counters[ind] = 0;
     imins[ind] = fftI+1;
     imaxs[ind] = -fftI-1;
     jmins[ind] = fftJ+1;
     jmaxs[ind] = -fftJ-1;
     kmins[ind] = fftK+1;
     kmaxs[ind] = -fftK-1;
     lmins[ind] = fftL+1;
     lmaxs[ind] = -fftL-1;
     mmins[ind] = fftM+1;
     mmaxs[ind] = -fftM-1;
    }

/* initiating the boundaries */

  boundaries[nbounds-1] = boundary;
  boundaries[0] = maxcoefficient;  
  for(ind=1; ind < nbounds-1; ind++)
    {
     boundaries[ind]=pow(BOUNDARY_STEP,n2-ind+1)/scale;
    }

/* statistical search */

  for (i=0; i < fftI; i++)
    {
     indI = fftJ * i;
     ci = (i > fftI/2) ? i-fftI : i;
     for (j=0; j < fftJ; j++)
       {
        indJ = fftK * ( j + indI );
        cj = (j > fftJ/2) ? j-fftJ : j;
        for (k=0; k < fftK; k++)          
	  {
           indK = fftL * ( k + indJ );
           ck = (k > fftK/2) ? k-fftK : k;
/* here we consider only the coefficients with the same parity of l and k */
/* for mode==2 l should be of the same parity as i+j+k */
/* for mode==3 l should be of the same parity as k */
           for (l=( (mode == 2) ? (i+j+k)%2 : k%2 ); l < fftL; l=l+2)
	     {
              indL = (fftM/2+1) * ( l + indK );
              cl = (l > fftL/2) ? l-fftL : l;
              for (m=0; m < fftM/2+1; m++)
		{
                 indM = m + indL;
                 cm=m;
/* any operations for data[indM] start here */

  cur = fabs(data[indM].re);

  for(ind=0; ind < nbounds; ind++)
    {
     if ( cur >= boundaries[ind] )
       {   
        for(cind=ind; cind < nbounds; cind++)
	  {
/* statistics about the distribution of significant coeffients */
           counters[cind]++;
/* statistics about positions of significant coeffients */
           if(imins[cind] > ci)
	     {
              imins[cind] = ci;
	     }
           if(imaxs[cind] < ci)
	     {
              imaxs[cind] = ci;
	     }
           if(jmins[cind] > cj)
	     {
              jmins[cind] = cj;
	     }
           if(jmaxs[cind] < cj)
	     {
              jmaxs[cind] = cj;
	     }
           if(kmins[cind] > ck)
	     {
              kmins[cind] = ck;
	     }
           if(kmaxs[cind] < ck)
	     {
              kmaxs[cind] = ck;
	     }    
           if(lmins[cind] > cl)
	     {
              lmins[cind] = cl;
	     }
           if(lmaxs[cind] < cl)
	     {
              lmaxs[cind] = cl;
	     }
           if(mmins[cind] > cm)
	     {
              mmins[cind] = cm;
	     }
           if(mmaxs[cind] < cm)
	     {
              mmaxs[cind] = cm;
	     }
	  }
        break;
       }
    }

/* any operations for data[indM] end here */
		}
	     }
	  }
       }
    }

/* statistical message about the number of significant coefficients */

/* the number of various pairs of k and l with the same parity is
   (fftK*fftL)/2+(fftK*fftL)%2 
*/

/* the number of different coefficient for m=0 and m=fftM/2 
   taking into account the symmetries for those values of m is
   (fftI/2+1)*(fftJ/2+1)*(fftK/2+1)*(fftL/2+1),
   but we do not use this symmetry in accordance with the form of 
      the cosine expansion.
*/

ncoeffs=fftI*fftJ*((fftK*fftL)/2+(fftK*fftL)%2)*(fftM/2+1);

printf("\nTotal number of computed coefficients:    "FORLINT"\n",ncoeffs);
printf("Total number of significant coefficients: "FORLINT"\n\n",counters[nbounds-1]);

for(ind=0; ind < nbounds; ind++)
  {
   printf(FORLINT" coefficients >= "FORDOUBLE" ("FORLINT":"FORLINT","FORLINT":"FORLINT","FORLINT":"FORLINT","FORLINT":"FORLINT","FORLINT":"FORLINT")\n",
          counters[ind],boundaries[ind]*scale,
          imins[ind],imaxs[ind],
          jmins[ind],jmaxs[ind],
          kmins[ind],kmaxs[ind],
          lmins[ind],lmaxs[ind],
          mmins[ind],mmaxs[ind]);
  }


/* finding lowest safe boundary */

ci=0;

for(ind=nbounds-1; ind >= 0; ind--)
  {
   if( safeq(imins[ind],imaxs[ind],fftI) &&
       safeq(jmins[ind],jmaxs[ind],fftJ) &&
       safeq(kmins[ind],kmaxs[ind],fftK) &&
       safeq(lmins[ind],lmaxs[ind],fftL) &&
       safeq(mmins[ind],mmaxs[ind],fftM) )
     {
      printf("\n\nThe lowest aliasing-safe boundary detected is "FORDOUBLE"\n",
             boundaries[ind]*scale);
      printf(FORLINT" coefficients are under the aliasing-safe boundary.\n",
             counters[ind]);

/* an aliasing-safe boundary was detected */
      ci=1;

      break;
     }
  }

if(ci==0)
  {
   printf("\n\nNo lowest aliasing-safe boundary is detected\n\n");
  }

/* freeing memory  */

  free(boundaries);
  free(counters);
  free(imins);
  free(imaxs);
  free(jmins);
  free(jmaxs);
  free(kmins);
  free(kmaxs);
  free(lmins);
  free(lmaxs);
  free(mmins);
  free(mmaxs);

}













